(*
 * Description:
 *
 * Copyright (c) 2000 by
 *  David Blei			blei@cs.berkeley.edu
 *  Chris Harrelson		chrishtr@cs.berkeley.edu
 *  Ranjit Jhala		jhala@cs.berkeley.edu
 *  Rupak Majumdar		rupak@cs.berkeley.edu
 *  George C. Necula	necula@cs.berkeley.edu
 *  Shree P. Rahul		sprahul@cs.berkeley.edu
 * 	Wes Weimer			weimer@cs.berkeley.edu
 *  Dror Weitz			dror@cs.berkeley.edu
 *   
 * All rights reserved.  Permission to use, copy, modify and distribute
 * this software for research purposes only is hereby granted, 
 * provided that the following conditions are met: 
 * 1.  Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 * 3. The name of the authors may not be used to endorse or promote products derived from 
 * this software without specific prior written permission. 
 *
 * DISCLAIMER:
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR IMPLIED 
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *)
(* Wes Weimer
 * CS 294-4, Spring '00
 *
 * General Tactical Satisfiability Procedure (Generic Inference Rules)
 * implements the signature in satproc.ml
 *
 * General Theory: we receive a message that indicates a goal. Check all of
 * our rules to see if they can handle it by using unification-based
 * pattern matchin against the goal expression. If it is a match and there
 * are no hypotheses, we prove it now. If it is a match and there are
 * hypotheses we use the substitution obtained from unification to
 * instantiate the hypotheses and submit them as subgoals with a proof
 * transformer (using the given inference rule) that turns a proof of the
 * subgoal into a proof of false. If the rule is more complicated (safeRd,
 * ofSel, etc) we may invoke arbitrary ML code to find "middle terms" to
 * flesh out the hypotheses. 
 *
 * GT never asserts equalities and does not keep track of anything for
 * snapshots/undos. 
 *)

module E = VampyreErrormsg
module I = Input
module P = Proof
module S = Satproc
open VampyrePretty
open Dag

(*
 * State Variables
 *)
let debug = ref false					(* set on command line *)
let match_debug = ref false		(* debug our internal pattern matching? *)
let rule_debug = ref false		(* debug our internal rule matching? *)
let gtid = ref ""

(*
 * The Internal "Language" For Inference Rules
 *
 * We use a bunch of helper functions to make writing inference rules down
 * here look as much like it does in axioms.lf as possible. We prepend most
 * of the terms with an underscore in order to avoid collisions with other
 * ML terms. 
 *)

(* general tactical pattern matching underscore *)
let __ v = (Input.Fun("__gt_underscore",[Input.Var(v)])) ;;
(* (__ "A") makes an externally universally quantified variable in an
 * inference rule definition. It must have the same value everywhere it
 * occurs. In particular, since we match based on the conclusion, we take
 * whatever value it was bound to when it matched the conclusion and
 * substitute in that value later when we emit the hypotheses as a new
 * subgoal. *)

(* types *)
let _int = Input.Fun("int",[]) ;;
let _bool = Input.Fun("bool",[]) ;;
let _exn = Input.Fun("exn",[]) ;;
let _ro = Input.Fun("ro",[]) ;;
let _rw = Input.Fun("rw",[]) ;;
let _ptr tau = (Input.Fun("ptr",[tau])) ;;
let _optptr tau = (Input.Fun("optptr", [tau])) ;; (* ptropt seems to be geared towards arrays.
						     For general pointers, we need optptr -- Rupak 
						   *)
let _ptropt tau o = (Input.Fun("ptropt",tau::[o])) ;;
let _array ro tau len = (Input.Fun("array",(ro :: tau :: [len]))) ;;
let _tuple2 tau1 tau2 = (Input.Fun("tuple2",(tau1 :: [tau2]))) ;;
let _tuple3 t1 t2 t3 = (Input.Fun("tuple3",(t1::t2::[t3]))) ;;
let _tuple4 t1 t2 t3 t4 = (Input.Fun("tuple4",(t1::t2::t3::[t4]))) ;;

(* bitwise operators *)
let _bitor x y = (Input.Fun ("bitOr", x::[y])) ;;
let _bitand x y = (Input.Fun ("bitAnd", x::[y])) ;;

(* Recursive types *)
let _mu t expt = (Input.Fun("mu",(t::[expt]))) ;;


(* typing predicates *)
let _of a b = (Input.Atom((Input.BFun("of",a :: [b])))) ;;
let _size4 e = (Input.Atom((Input.BFun("size4",[e])))) ;;
let _saferd4 m a = (Input.Atom((Input.BFun("saferd4",m::[a])))) ;;
let _safewr4 m a e = (Input.Atom((Input.BFun("safewr4",(m::a::[e]))))) ;;
let _fieldOf t1 sizet off t2 = (Input.Atom((Input.BFun("fieldOf",(t1::sizet::off::[t2]))))) ;;
let _arrIdx i ol e ow = (Input.Atom((Input.BFun("arrIdx",(i::ol::e::[ow]))))) ;;

(* generic predicates *)
let _neq a b = (Input.Atom((Input.Neq(a,b)))) ;;
let _geq a b = (Input.Atom((Input.Geq(a,b)))) ;;
let _gt a b = (Input.Atom((Input.Gt(a,b)))) ;;
let _lt a b = (Input.Atom((Input.Lt(a,b)))) ;;
let _eq  a b = (Input.Atom((Input.Eq(a,b)))) ;;

(* Boolean equality *)
let _beq a b = (Input.Atom (Input.Beq(a,b))) ;;

(* memory expressions *)
let _sel4 m a = (Input.Fun("sel4",m :: [a])) ;;
let _upd4 m a e = (Input.Fun("upd4",m :: a :: [e])) ;;

(* generic expressions *)
let _true = Input.True ;;
let _false = Input.False ;;
let _num i = (Input.Int(i)) ;;
let _mod i e = (Input.Fun("__gt_modulo",i :: [e])) ;;
let _div i e = (Input.Fun("__gt_div",i :: [e])) ;;
let _plus a b = (Input.Plus(a,b)) ;;
let _minus a b = (Input.Minus(a,b)) ;;
let _times a b = (Input.Times(a,b)) ;;

(*
 * Generic Inference Rules
 *)
type rule = {
    ident   : S.ident ;           (* the name of this inference rule *)
    hyp	    : Input.pred list ;	  (* its hypotheses *)
    conc    : Input.pred ;	  (* its conclusion *)
    exp     : Input.exp list ; 	  (* explicit proof arguments *)
}
      
(*
 * Typing Rules from axioms.lf
 *)
let ruleList = [
        { ident = "bitOrAnd" ;
          hyp = (_eq (__ "Z") (_bitor (__ "X") (__ "C"))) :: [] ;
          conc = _neq (_bitand (__ "Z") (__ "C")) (_num 0) ;
          exp = [] ; };
(*        { ident = "bit" ;
	  hyp = (_eq (__ "Z") (_bitor (__ "X") (__ "C"))) :: [ _eq (_bitand (__ "Z") (__ "C")) (_num 0) ] ;
	  conc = _false ;
	  exp = [] ;
	} ;
*)
(*
        { ident = "ofExn0" ; hyp = []; conc = _of (_num 0) _exn ; exp = []; } ;
	{ ident = "ofExn1" ; hyp = []; conc = _of (_num 1) _exn ; exp = []; } ;
	{ ident = "ofExn2" ; hyp = []; conc = _of (_num 2) _exn ; exp = []; } ;
	{ ident = "ofExn3" ; hyp = []; conc = _of (_num 3) _exn ; exp = []; } ;
	{ ident = "ofExn4" ; hyp = []; conc = _of (_num 4) _exn ; exp = []; } ;
	{ ident = "ofExn5" ; hyp = []; conc = _of (_num 5) _exn ; exp = []; } ;
	{ ident = "ofIntAny" ; hyp = []; conc = _of (__ "A") _int ; 
		exp = [(__ "A")] ; } ;
	{ ident = "ofBool0" ; hyp = []; conc = _of (_num 0) _bool ; exp = [];} ;
	{ ident = "ofBool1" ; hyp = []; conc = _of (_num 1) _bool ; exp = [];} ;
	{ ident = "szint" ; hyp = []; conc = _size4 _int ; exp = [];} ;
	{ ident = "szbool" ; hyp = []; conc = _size4 _bool ; exp = []; } ;
	{ ident = "szarr" ; hyp = []; 
		conc = _size4 (_array (__ "RO") (__ "T") (__ "LEN")) ; 
		exp = [(__ "RO");(__ "T");(__ "LEN")]; } ;
	{ ident = "szptropt" ; hyp = [];
	  conc = _size4 (_ptropt (__ "T") (__ "LEN")) ; 
		exp = [(__ "T");(__ "LEN")]; } ;
	{ ident = "szoptptr" ; hyp = [];
	  conc = _size4 (_optptr (__ "T")) ; 
		exp = [(__ "T")]; } ;
	{ ident = "fo10" ; hyp = [ _size4 (__ "T") ] ; 
	  conc = _fieldOf (__ "T") (_num 4) (_num 0) (__ "T") ; exp = []; } ;
	{ ident = "fo20" ; hyp = _size4 (__ "T1") :: [_size4 (__ "T2") ] ;
	  conc = _fieldOf (_tuple2 (__ "T1") (__ "T2")) 
										(_num 8) (_num 0) (__ "T1") ; exp = [] ; } ;
	{ ident = "fo21" ; hyp = _size4 (__ "T1") :: [_size4 (__ "T2") ] ;
	  conc = _fieldOf (_tuple2 (__ "T1") (__ "T2")) 
										(_num 8) (_num 4) (__ "T2") ; exp = [] ; } ;
	{ ident = "fo30" ; 
	  hyp = _size4 (__ "T1") :: _size4 (__ "T2") :: [_size4 (__ "T3") ] ;
	  conc = _fieldOf (_tuple3 (__ "T1") (__ "T2") (__ "T3")) 
										(_num 12) (_num 0) (__ "T1") ; exp = []; } ;
	{ ident = "fo31" ; 
	  hyp = _size4 (__ "T1") :: _size4 (__ "T2") :: [_size4 (__ "T3") ] ;
	  conc = _fieldOf (_tuple3 (__ "T1") (__ "T2") (__ "T3")) 
										(_num 12) (_num 4) (__ "T2") ; exp = []; } ;
	{ ident = "fo32" ; 
	  hyp = _size4 (__ "T1") :: _size4 (__ "T2") :: [_size4 (__ "T3") ] ;
	  conc = _fieldOf (_tuple3 (__ "T1") (__ "T2") (__ "T3")) 
										(_num 12) (_num 8) (__ "T3") ; exp = []; } ;
	{ ident = "fo40" ; hyp = _size4 (__ "T1") :: _size4 (__ "T2") :: 
													 _size4 (__ "T3") :: [_size4 (__ "T4") ] ;
	  conc = _fieldOf (_tuple4 (__ "T1") (__ "T2") (__ "T3") (__ "T4") ) 
										(_num 16) (_num 0) (__ "T1") ; exp = []; } ;
	{ ident = "fo41" ; hyp = _size4 (__ "T1") :: _size4 (__ "T2") :: 
													 _size4 (__ "T3") :: [_size4 (__ "T4") ] ;
	  conc = _fieldOf (_tuple4 (__ "T1") (__ "T2") (__ "T3") (__ "T4") ) 
										(_num 16) (_num 4) (__ "T2") ; exp = []; } ;
	{ ident = "fo42" ; hyp = _size4 (__ "T1") :: _size4 (__ "T2") :: 
													 _size4 (__ "T3") :: [_size4 (__ "T4") ] ;
	  conc = _fieldOf (_tuple4 (__ "T1") (__ "T2") (__ "T3") (__ "T4") ) 
										(_num 16) (_num 8) (__ "T3") ; exp = []; } ;
	{ ident = "fo43" ; hyp = _size4 (__ "T1") :: _size4 (__ "T2") :: 
													 _size4 (__ "T3") :: [_size4 (__ "T4") ] ;
	  conc = _fieldOf (_tuple4 (__ "T1") (__ "T2") (__ "T3") (__ "T4") ) 
										(_num 16) (_num 12) (__ "T4") ; exp = []; } ;
  { ident = "aIdx" ; 
	  hyp = _geq (__ "I") (_num 0) ::
				  _geq (__ "Len") (__ "I") ::
				[ _eq (_mod (__ "I") (__ "ELen") ) (__ "Off") ] ;
		conc = _arrIdx (__ "I") (__ "Len") (__ "ELen") (__ "Off") ; 
		exp = [(__ "ELen") ; (__ "Off")] ; } ;
	{ ident = "safeRd" ;
		hyp = _eq (__ "ADDR") (_plus (__ "A") (__ "I")) ::
					_fieldOf (__ "T") (__ "SIZET") (__ "OFF") (__ "T'") ::
					_arrIdx (__ "I") (__ "L") (__ "SIZET") (__ "OFF") ::
				[ _of (__ "A") (_array (__ "RO") (__ "T") (__ "L")) ] ;
		conc = _saferd4 (__ "M") (__ "ADDR") ; exp = [(__ "M")] ; } ; 
	{ ident = "safeWr" ; 
		hyp = _of (__ "E") (__ "T'") ::
					_eq (__ "ADDR") (_plus (__ "A") (__ "I")) ::
					_fieldOf (__ "T") (__ "SIZET") (__ "OFF") (__ "T'") ::
					_arrIdx (__ "I") (__ "L") (__ "SIZET") (__ "OFF") ::
				[ _of (__ "A") (_array _rw (__ "T") (__ "L")) ] ;
		conc = _safewr4 (__ "M") (__ "ADDR") (__ "E") ; exp = [(__ "M")] ;} ; 
	{ ident = "ofSel" ;
		hyp = _eq (__ "ADDR") (_plus (__ "A") (__ "I")) ::
					_fieldOf (__ "T") (__ "SIZET") (__ "OFF") (__ "T'") ::
					_arrIdx (__ "I") (__ "L") (__ "SIZET") (__ "OFF") ::
				[ _of (__ "A") (_array (__ "RO") (__ "T") (__ "L")) ] ;
		conc = _of (_sel4 (__ "M") (__ "ADDR")) (__ "T'") ; exp = [(__ "M")] ;} ;
	(* ofSel seems to handle ofSelId well enough *)
	(* { ident = "ofSelId" ;
		hyp = _fieldOf (__ "T") (__ "SIZET") (__ "OFF") (__ "T'") ::
					_arrIdx (__ "I") (__ "L") (__ "SIZET") (__ "OFF") ::
				[ _of (__ "A") (_array (__ "RO") (__ "T") (__ "L")) ] ;
		conc = _of (_sel4 (__ "M") (_plus (__ "A") (__ "I"))) (__ "T'") ; } ; *)
	{ ident = "geqSel" ;
		hyp = _eq (__ "M") (__ "M2") ::
		      _eq (_sel4 (__ "M") (__ "ADDR")) (__ "OTHER") ::
				[ _geq (__ "GOAL") (__ "OTHER") ] ;
		conc = _geq (__ "GOAL") (_sel4 (__ "M") (__ "ADDR")) ; exp = [] ; } ; 
        { ident = "ofPtrOpt" ;
                hyp =	_neq (__ "A") (_num 0) ::
                           [ _of (__ "A") (_ptropt (__ "T") (__ "LEN")) ] ;
                conc = 	_of (__ "A") (_array _rw (__ "T") (__ "LEN")) ; exp = [] ; } ;


	{ ident = "of0POpt" ;
	  hyp = [] ; conc = _of (_num 0) (_ptropt (__ "T") (__ "LEN")) ; 
		exp = [(__ "T");(__ "LEN")] ; } ;

  
      (* Type Congruence *)
(*
        { ident = "cong" ;
          hyp = (_of (__ "E") (__ "T")) :: [ _eq (__ "E") (__ "E2") ] ;
	  conc = _of (__ "E2") (__ "T") ;
	  exp = [] ;
	} ;
*)


  (* Rupak Addition:
            Rule for optptr
               A : optptr (T)    A != 0
             ----------------------------
                   A : ptr (T)
            Corresponding axiom:
             ofPtr : {A:exp}{T:exp} % End of implicit arguments
                     pf(<> A 0) ->
                     pf(of A (optptr T)) ->
                     pf(of A (ptr T))
   *)
        { ident = "ofPtr" ;
          hyp   = _neq (__ "A") (_num 0) ::
                  [ _of (__ "A") (_optptr (__ "T")) ] ;
	  conc =  _of (__ "A") (_ptr (__ "T")) ; 
	  exp = [] ; 
	} ;

   (*  Null: 0 : optptr(T) *)
        { ident = "null" ;
	  hyp = [] ;
	  conc = _of (_num 0) (_optptr (__ "T")) ;
	  exp = [(__ "T")] ; } ;
   (* Rupak Addition:
           Rule optptri
                    A : ptr(T)
              ------------------------------
                   A : optptr(T)
           Corresponding axiom
           ofPtri : {A: exp} {T: exp} % End of implicit arguments
                    pf (of A (ptr T)) ->
                    pf (of A (optptr T))
    *)
        { ident = "optPtri" ;
          hyp   = [ _of (__ "A") (_ptr (__ "T")) ] ;
	  conc =  _of (__ "A") (_optptr (__ "T")) ; 
	  exp = [] ;
	} ;


	{ ident = "ofArrPOpt" ;
	  hyp = [ _of (__ "A") (_array _rw (__ "T") (__ "LEN")) ] ;
		conc = _of (__ "A") (_ptropt (__ "T") (__ "LEN")) ; exp = []; } ;

        (* Type Equality *)
        { ident = "=~of" ;
	  hyp = _eq (__ "E1") (__ "E2"):: [ _eq (__ "T1") (__ "T2") ] ;
	  conc = _eq (Input.Fun ("of", (__ "E1")::[(__ "T1")])) 
	              (Input.Fun ("of", (__ "E2")::[(__ "T2")])) ; 
	  exp = [] ;
	} ;


	{ ident = "mc0" ;
	  hyp = [ _eq (__ "A'") (__ "A") ] ;
	  conc = _eq  (_sel4 (_upd4 (__ "M") (__ "A") (__ "E")) (__ "A'")) 
	    (__ "E") ; 
	  exp = [(__ "M");(__ "E")] ; } ;
	{ ident = "mc1" ;
	  hyp = [ _neq (__ "A'") (__ "A") ] ;
		conc = _eq  (_sel4 (_upd4 (__ "M") (__ "A") (__ "E")) (__ "A'"))
								(_sel4 (__ "M") (__ "A'")) ; 
		exp = [(__ "M");(__ "E")] ; } ;

	(* non-linear axioms for bidimensional array access *)
	{ ident = "bidmi0" ; exp = [] ;
	  hyp =	_gt (_minus (__ "LY") (__ "I")) (_num 0) ::
					_geq (__ "I") (_num 0) ::
				[ _geq (__ "J") (_num 0) ] ;
		conc = _geq (_plus (__ "I") (_times (__ "LY") (__ "J"))) (_num 0) ; } ;
	{ ident = "bidmi0a" ; exp = [] ;
	  hyp =	_geq (__ "LY") (_num 0) ::
					_geq (__ "I") (_num 0) ::
				[ _geq (__ "J") (_num 0) ] ;
		conc = _geq (_plus (__ "I") (_times (__ "J") (__ "LY"))) (_num 0) ; } ;
	{ ident = "bidim1" ; exp = [] ;
		hyp = _gt (_minus (__ "LX") (__ "J")) (_num 0) ::
					_gt (_minus (__ "LY") (__ "I")) (_num 0) ::
				[ _geq (__ "I") (_num 0) ] ;
		conc = _geq (_minus (_times (_num 4) (_times (__ "LX") (__ "LY")))
												(_num 4))
								(_times (_plus (__ "I") (_times (__ "LY") (__ "J")))
												(_num 4)) ; } ;
	{ ident = "bidim2" ; exp = [] ;
		hyp = _lt (_minus (__ "X") (__ "LX")) (_num 0) ::
					_gt (_minus (__ "LY") (__ "Y")) (_num 0) ::
					_eq (__ "LX'") (_times (__ "LX") (_num 4)) ::
				[ _geq (__ "LX'") (_num 0) ] ;
		conc = _geq (_minus (_times (__ "LX") (__ "LX'")) (_num 4))
								(_times (_plus (__ "X") (_times (__ "Y") (__ "LX"))) (_num 4))
		; } ; 
	{ ident = "bidim3" ; exp = [] ;
	  hyp =  _geq (__ "X") (_num 0) ::
					 _geq (__ "Y") (_num 0) ::
				[  _geq (__ "Z") (_num 0) ]  ;
		conc = _geq (_times (_plus (__ "X") (_times (__ "Y") (__ "Z"))) (_num 4)) 
								(_num 0) ; } ; 
	(* modular arithmetic *)
	{ ident = "modGte0" ; exp = [(__ "X");(__ "Y")] ;
	  hyp = [ _geq (__ "Z") (_num 0) ] ;
		conc = _geq (_times (_mod (__ "X") (__ "Y")) (__ "Z")) (_num 0) ; } ;
  { ident = "modCt" ; 
    hyp = [] ; 
		conc = _eq (_mod (__ "E") (__ "N")) (__ "RES") ; 
		exp = [(__ "E");(__ "N");(__ "RES")] ; } ;
	{ ident = "divCt" ;
	  hyp = [] ; (* TODO: N>0, E div N = RES *)
		conc = _eq (_div (__ "E") (__ "N")) (__ "RES") ; 
		exp = [(__ "N");(__ "E");(__ "RES")] ; } ;
	{ ident = "modLeq" ;
	  hyp = [] ; (* TODO: N>0, Nm1 = N-1 *)
		conc = _geq (__ "Nm1") (_mod (__ "E") (__ "N")) ; 
		exp = [(__ "Nm1");(__ "E");(__ "N")] ; } ;
	{ ident = "modGeq" ;
	  hyp = [] ; (* N>0 *)
		conc = _geq (_mod (__ "E") (__ "N")) (_num 0) ; 
		exp = [(__ "N");(__ "E")] ; } ;
	{ ident = "mod*+" ;
	  hyp = [] ; (* n mod DV = N', F mod DV = 0 *)
		conc = _eq (_mod (_plus (__ "N") (_times (__ "E") (__ "F"))) (__ "DV"))
							 (__ "N'") ; 
		exp = [(__ "E");(__ "N'");(__ "N");(__ "F");(__ "DV")]; } ; 
	{ ident = "mod*" ;
	  hyp = [] ; (* F mod DV = 0 *)
		conc = _eq (_mod (_times (__ "E") (__ "F")) (__ "DV")) (_num 0) ; 
		exp = [(__ "E");(__ "F");(__ "DV")] ; } ;
 	{ ident = "div+" ;
		hyp = [] ; (* F mod DV = 0, N div DV = N', F div DV = F' *)
		conc = _eq (_div (_plus (__ "N") (_times (__ "E") (__ "F"))) (__ "DV"))
							 (_plus (__ "N'") (_times (__ "E") (__ "F'"))) ; 
		exp = [(__ "E");(__ "F'");(__ "N");(__ "N'");(__ "F");(__"DV")]; } 
	*)
  ] 
 ;;

type specialSubgoal = {
    specialId : S.ident ;           (* the name of this inference rule *)
    makeSubgoals : rule -> Satproc.message -> (string * I.exp) list -> unit ; 
      (* make the subgoal[s] *)
}

(* 
 * Inference Rule Pattern Matching
 *
 * We do our own Robinson-style unification to pattern match inference
 * rules (which can contain unbound variables like (__ "A")) with I.exp's 
 * (goals we received from the outside world via "assume") in order to see
 * if the rule is applicable. If the conclusion matches the goal node
 * we just received, it is applicable. In that case, we'll use the
 * substituion given by the pattern matching and apply it to the
 * hypotheses: they then become the new subgoals. 
 *)

exception No_Match	(* internal: it doesn't unify *)

(* Home-brew pattern matching main entry point. gt_rule is an Input.pred
 * and assumed_thing is a [goal] I.pred. This is just standard unification
 * -- it doesn't even have the dreaded occurs check because variables can
 * only appear on the gt_rule side. 
 *
 * Within these functions, b* is the built-up environment, a list of
 * (var,exp) pairs for gt_rule variables. g* is part of the gt_rule and
 * a* is part of the assumed term. 
 *)
let matches gt_rule assumed_thing =
  let rec matches_atom b1 g a =
    if (!match_debug) then ignore 
	(printf "GT: matches_atom current rule = @[%a@]\nassumed term = @[%a@]\n" I.dp (I.Atom(g))
	   I.dp (I.Atom(a))) ;
    match (g,a) with 
    | ( (I.Beq(a1,a2)), (I.Beq(a3,a4)) ) ->
	let b2 = matches_atom b1 a1 a3 in
	let b3 = matches_atom b2 a2 a4 in b3
    | ( (I.Eq(e1,e2)), (I.Eq(e3,e4)) ) ->
	let b2 = matches_exp b1 e1 e3 in
	let b3 = matches_exp b2 e2 e4 in b3
    | ( (I.Neq(e1,e2)), (I.Neq(e3,e4)) ) ->
	let b2 = matches_exp b1 e1 e3 in
	let b3 = matches_exp b2 e2 e4 in b3
    | ( (I.Geq(e1,e2)), (I.Geq(e3,e4)) ) ->
	let b2 = matches_exp b1 e1 e3 in
	let b3 = matches_exp b2 e2 e4 in b3
    | ( (I.Leq(e1,e2)), (I.Leq(e3,e4)) ) ->
	let b2 = matches_exp b1 e1 e3 in
	let b3 = matches_exp b2 e2 e4 in b3
    | ( (I.Gt(e1,e2)), (I.Gt(e3,e4)) ) ->
	let b2 = matches_exp b1 e1 e3 in
	let b3 = matches_exp b2 e2 e4 in b3
    | ( (I.Lt(e1,e2)), (I.Lt(e3,e4)) ) ->
	let b2 = matches_exp b1 e1 e3 in
	let b3 = matches_exp b2 e2 e4 in b3
    | ( (I.BFun(s1,l1)), (I.BFun(s2,l2)) ) when (s1 = s2) ->
	let b_final = List.fold_left2 matches_exp b1 l1 l2 in
	b_final
    | _ -> raise No_Match
  and matches_exp b1 g a =
    if (!match_debug) then ignore 
	(printf "GT: matches_exp @[%a @!%a@]\n" I.de g I.de a) ;
    match (g,a) with
      ( (I.Var(v1)), (I.Var(v2)) ) when ((v1 = v2) or 
					 (Dag.knownEq (Dag.internExp g) (Dag.internExp a))) -> b1
    | ( (I.Int(i1)), (I.Int(i2)) ) when (i1 = i2) -> b1
    | ( (I.Plus(e1,e2)), (I.Plus(e3,e4)) ) ->
	let b2 = matches_exp b1 e1 e3 in
	let b3 = matches_exp b2 e2 e4 in b3
    | ( (I.Minus(e1,e2)), (I.Minus(e3,e4)) ) ->
	let b2 = matches_exp b1 e1 e3 in
	let b3 = matches_exp b2 e2 e4 in b3
    | ( (I.Times(e1,e2)), (I.Times(e3,e4)) ) ->
	let b2 = matches_exp b1 e1 e3 in
	let b3 = matches_exp b2 e2 e4 in b3
    | ( (I.Div(i1,e1)), (I.Div(i2,e2)) ) when (i1 = i2) ->
	let b2 = matches_exp b1 e1 e2 in b2
    | ( (I.Mod(i1,e1)), (I.Mod(i2,e2)) ) when (i1 = i2) ->
	let b2 = matches_exp b1 e1 e2 in b2
    | ( (I.Floor(i1,e1)), (I.Floor(i2,e2)) ) when (i1 = i2) ->
	let b2 = matches_exp b1 e1 e2 in b2
    | ( (I.Ceil(i1,e1)), (I.Ceil(i2,e2)) ) when (i1 = i2) ->
	let b2 = matches_exp b1 e1 e2 in b2
    | ( (I.Fun(s1,l1)), (I.Fun(s2,l2)) ) when (s1 = s2) ->
	let b_final = List.fold_left2 matches_exp b1 l1 l2 in
	b_final
		(* special handling for modulo because Input.Mod(i,e) requires that
		   * "i" be an int, so we can't put (__ "I") there directly. Instead,
		   * we let the user write (_mod (__ "I") ...) in the rules, and we
		   * delay the unification until here. *)
    | ( (I.Fun("__gt_modulo", i1 :: [e1])), (I.Mod(i2,e2)) ) ->
	let b2 = matches_exp b1 i1 (I.Int(i2)) in
	let b3 = matches_exp b2 e1 e2 in b3
    | ( (I.Fun("__gt_div", i1 :: [e1])), (I.Div(i2,e2)) ) ->
	let b2 = matches_exp b1 i1 (I.Int(i2)) in
	let b3 = matches_exp b2 e1 e2 in b3
               (* this is the bit where we actually bind pattern variables to
                  * expression values. if everything is consistent, we prepend the
                  * new binding onto the list. *)
    | ( (I.Fun("__gt_underscore", [I.Var(v1)])), an_exp ) -> 
	if (List.mem_assoc v1 b1) then (* we already have a binding for v1! *)
		      (* so it better be consistent! *)
	  let b2 = matches_exp b1 (List.assoc v1 b1) an_exp in b1
	else (* this is the first time we've seen v1 *)
	  ((v1,an_exp) :: b1) (* so bind it to an_exp *)
	    
    | (_,_) -> raise No_Match
  in match gt_rule with
    I.Atom(a) -> matches_atom [] a (Subgoal.inputAtomOfDagNode assumed_thing) 
  | _ -> E.s (E.bug "GT cannot handle non I.Atom based predicates")
	;;

(*
 * Subgoal Generation
 *
 * We use the substitution given by the pattern matcher, apply it to the
 * hypotheses, and spit the result out as the subgoal. 
 *)

(* subgoal generation with substitution: the only real work here is
 * replacing underscore'd variables with their bound values. *)
let rec subGoalOfPred np s = 
	match np with
	  I.All(v,p) -> (I.All(v,(subGoalOfPred p s)))
	| I.Exist(v,p) -> (I.Exist(v,(subGoalOfPred p s)))
	| I.And(p1,p2) -> (I.And((subGoalOfPred p1 s),(subGoalOfPred p2 s)))
	| I.Or(p1,p2) -> (I.Or((subGoalOfPred p1 s),(subGoalOfPred p2 s)))
	| I.Imp(p1,p2) -> (I.Imp((subGoalOfPred p1 s),(subGoalOfPred p2 s)))
	| I.Comment(str,p) -> (I.Comment(str,(subGoalOfPred p s)))
	| I.Atom(a) -> (I.Atom((subGoalOfAtom a s)))
	|	_ -> np
and subGoalOfAtom ba s =
	match ba with
		I.Eq(a,b) -> (I.Eq((subGoalOfExp a s),(subGoalOfExp b s)))
	|	I.Gt(a,b) -> (I.Gt((subGoalOfExp a s),(subGoalOfExp b s)))
	|	I.Lt(a,b) -> (I.Lt((subGoalOfExp a s),(subGoalOfExp b s)))
	|	I.Neq(a,b) -> (I.Neq((subGoalOfExp a s),(subGoalOfExp b s)))
	|	I.Geq(a,b) -> (I.Geq((subGoalOfExp a s),(subGoalOfExp b s)))
	|	I.Leq(a,b) -> (I.Leq((subGoalOfExp a s),(subGoalOfExp b s)))
	| I.NotBFun(str,l) -> let myfun a = (subGoalOfExp a s) in
			(I.NotBFun(str,(List.map myfun l)))
	|	I.BFun(str,l) -> let myfun a = (subGoalOfExp a s) in
			(I.BFun(str,(List.map myfun l)))
	| I.Beq(a,b) -> E.s (E.bug "subGoalOfAtom encountered a Beq, call Jhala!")
and subGoalOfExp be s =
	match be with
	  (I.Fun("__gt_underscore", [I.Var(v)])) -> 
	    List.assoc v s (* this is it! look up the value in the environment *)
	| (I.Fun("__gt_modulo", [i ; e])) -> 
	    let s_i = (subGoalOfExp i s) in
	    let s_e = (Sup_inf.simplify (subGoalOfExp e s)) in (* integer *)
	    let the_int = (match s_e with
	      I.Int(ii) -> ii
	    | _ -> E.s (E.bug "subGoalOfAtom: mod problems: %a %a\n" I.de s_i
			  I.de s_e)) in
	                    (I.Mod(the_int,s_i))
        | I.Plus(a,b) -> (I.Plus((subGoalOfExp a s),(subGoalOfExp b s)))
	| I.Minus(a,b) -> (I.Minus((subGoalOfExp a s),(subGoalOfExp b s)))
	| I.Times(a,b) -> (I.Times((subGoalOfExp a s),(subGoalOfExp b s)))
	| I.Div(i,a) -> (I.Div(i,(subGoalOfExp a s)))
	| I.Mod(i,a) -> (I.Mod(i,(subGoalOfExp a s)))
	| I.Floor(i,a) -> (I.Floor(i,(subGoalOfExp a s)))
	| I.Ceil(i,a) -> (I.Ceil(i,(subGoalOfExp a s)))
	| I.Fun(str,l) -> let myfun a = (subGoalOfExp a s) in 
	            (I.Fun(str,(List.map myfun l)))
	| _ -> be
(* this is the entry point to these procedures: it returns a new
 * Subgoal.subgoal based on the given rule and substitution. *)
and makeSubGoal rule goal_msg s prio = 
  let my_sub_pred_fun a = (subGoalOfPred a s) in
  let my_box_fun a = (Proof.P(a)) in 
  let my_sub_exp_fun a = (Proof.E((subGoalOfExp a s))) in 
  let newSubGoal = {
    Subgoal.owner = !gtid; 		(* we ("tacticals") made this subgoal *)
    Subgoal.priority = prio; 		(* default priority *)
    Subgoal.goal = (List.map my_sub_pred_fun rule.hyp);
    Subgoal.proof_gen = (function hyp_proof_list -> 
      (P.G("not_FI",[(P.P(P.G(rule.ident, 
			      (List.map my_sub_exp_fun rule.exp) @
			      (List.map my_box_fun hyp_proof_list)))) ;
		      (P.P(!Satproc.proveIt goal_msg))])))
  } in
  newSubGoal
;;

(*
 * Special Subgoal Creation
 *
 * For rules that are not well-suited to backwards chaining, we must invent
 * intermediate terms (to be used in subgoals). Also, for rules with
 * sideconditions, we must check to see if they work before proving them. 
 *)

(* sizeOfType of an inputExp returns an Input.Exp describing the size of the
 * type. Currently incomplete. *)
let rec numSizeOfType dn = 
	match dn with
		I.Fun("int",[]) -> 4
	| I.Fun("bool",[]) -> 4
	| I.Fun("byte",[]) -> 1
	| I.Fun("exn",[]) -> 4
	| I.Fun("ptr",[]) -> 4
	| I.Fun("optptr",[tau]) -> 4	
	| I.Fun("ptropt",[tau;I.Int l]) -> 4 + l
	| I.Fun("array",[rw;tau;I.Int l]) -> 4 + l
	| I.Fun("tuple2",args) -> (numSizeOfType (List.nth args 0)) +
												 (numSizeOfType (List.nth args 1))
	| I.Fun("tuple3",args) -> (numSizeOfType (List.nth args 0)) +
												 (numSizeOfType (List.nth args 1)) +
												 (numSizeOfType (List.nth args 2))
	| I.Fun("tuple4",args) -> (numSizeOfType (List.nth args 0)) +
												 (numSizeOfType (List.nth args 1)) +
												 (numSizeOfType (List.nth args 2)) +
												 (numSizeOfType (List.nth args 3))
	| _ -> E.s (E.bug "GT: cannot determine size of type %a" I.de dn)
;;
let sizeOfType dn = I.Int(numSizeOfType dn) ;;
		
(* returns true if the I.Exp addr contains a mention of the variable "arry"
 * anywhere within its recursive structure. *)
let rec addrContainsArray addr arry =
	match addr with
		I.Var(v) -> (v = arry)
	| I.Fun(v,[]) -> (v = arry)
	| I.Fun("sel4",[m;a]) -> (addrContainsArray a arry)
	| I.Plus(a,b) -> (addrContainsArray a arry) or (addrContainsArray b arry)
	| I.Minus(a,b) -> (addrContainsArray a arry) or (addrContainsArray b arry)
	| I.Times(a,b) -> (addrContainsArray a arry) or (addrContainsArray b arry)
	| I.Div(i,b) -> (addrContainsArray b arry)
	| I.Mod(i,b) -> (addrContainsArray b arry)
	| I.Floor(i,b) -> (addrContainsArray b arry)
	| I.Ceil(i,b) -> (addrContainsArray b arry)
	| _ -> false
;; 

(* 
 * Removes the array from the address expression to get the index. 
 *) 
let rec removeArrayFromAddr addr arry =
	match addr with
		I.Plus(I.Var(arry),e) -> e
	| I.Plus(e,I.Var(arry)) -> e
	| I.Minus(e,I.Var(arry)) -> e
	| I.Plus(a,b) -> I.Plus(removeArrayFromAddr a arry,b)
	| I.Var(a) when a = arry -> I.Int 0
	| I.Fun(v,[]) when v = arry -> I.Int 0
	| I.Fun("sel4",[m;e]) -> (I.Fun("sel4",[m;removeArrayFromAddr e arry]))
	| I.Minus(a,b) when (b <> I.Var(arry)) -> 
			I.Minus(removeArrayFromAddr a arry,b)
	| _ -> E.s (E.bug "GT: removeArrayFromAddr %s in %a\n" arry I.de addr)
;;

(*
 * Checks to see if "modCt N E RES" == "(E mod N) = RES" == 
 * "(= (mod E N) RES)" is provable. 
 * N == ELen	==  SIZET
 * E == I 		==  I
 * RES == Off ==  OFF
 *
 * Basically, we check a bunch of special cases that came up in our
 * examples.
 *)
let modCheck nn ee rres =
	match (nn,ee,rres) with
		(I.Int n),(I.Int e),(I.Int r) ->
			if (n > 0) & (r >= 0) & (n > r) & ((e mod n) = r) then true
		  else false
	| (I.Int n),(I.Times(x,y)),(I.Int r) when (y = I.Int n) & (r = 0) -> true
	| (I.Int n),(I.Plus(I.Times(x,y),z)),(I.Int r) when
		(y = I.Int n) & (z = I.Int r) -> true
	| (I.Int n),(I.Plus(I.Times(x,y),z)),(I.Int 0) when
		(y = I.Int n) & (z = I.Int n) -> true
	| (I.Int n),(I.Plus(I.Times(I.Minus(I.Times(x,I.Int n1),y),I.Int n2),I.Int n3)),(I.Int n4) when (n1-1)*n2 = n && n3=n4 -> true
	| (I.Int n),(I.Times(I.Minus(I.Times(x,I.Int n1),y),I.Int n2)),(I.Int 0) 
		when (n1-1)*n2 = n -> true
	| (I.Int n),(I.Plus(I.Int x,(I.Times(y,I.Int z)))),(I.Int r) when
		(x = r) & (n = z) -> true
	(* modCheck 4 ((X9_10 * 4) - 4) 0 *)
	| (I.Int n),(I.Minus(I.Times(x,I.Int y),I.Int z)),(I.Int 0) when
		(n = y) & (n = z) -> true
	| _ -> false
;;

(*
 * Simplifies arithmetic expressions (like array indices). cf.
 * Sup_inf.simplify. 
 *)
let rec gtSimplify i =
	match i with
		I.Plus(I.Int 0,a) -> gtSimplify a
	| I.Plus(a,I.Int 0) -> gtSimplify a
	| I.Plus(a,b) -> I.Plus(gtSimplify a,gtSimplify b)
	| I.Minus(a,I.Int 0) -> gtSimplify a
	| I.Minus(a,b) -> I.Minus(gtSimplify a,gtSimplify b)
	| I.Times(a,b) -> I.Times(gtSimplify a,gtSimplify b)
	| _ -> i
		

(* find an bitwise-or to serve as the intermediate step for bitOrAnd,
 * Makes a new substitution based on that bitOr and then submits it as
 * a subgoal.
 *)
let findBitOr r msg s =
  (if (!debug) then  
    ignore (printf "GT: findBit subgoal creation for %a\n" Satproc.dm msg)) ;
  printf "GT: Association of Z is %a\n" I.de (List.assoc "Z" s) ;
  for i = 1 to (Dag.dagSize () ) do 
    let node_i = Dag.findDag i in 
    if (node_i.op = Dag.Fun ("bitOr")) then (
      match (node_i.args) with
	(* strategy: consider every bitOr so far *)
      |	[a;b] ->
	  (if (!debug) then ignore (printf "findBitOr: valid assignment!\n"));
	  if(!debug) then ignore (
	    printf "GT: matched %a = %a\n" I.de (Subgoal.inputExpOfDagNode a) I.de (Subgoal.inputExpOfDagNode b)) ;
	  let s_X = (Subgoal.inputExpOfDagNode a) in
	  let s_C = (Subgoal.inputExpOfDagNode b) in
	  let new_s = ("X",s_X) :: ("C", s_C) :: s in
	  Subgoal.add (makeSubGoal r msg new_s 1)
      | _ -> () 
	    )
    else ()
  done
;;

(*
 * The big procedure: finds an array (or ptropt) to serve as the
 * intermediate step for saferd, safewr, ofsel, etc. Makes a new
 * substitution based on that array and then submits it as a subgoal. 
 *)
let findArray r msg s =
	let did_something = ref false in 
	(if (!debug) then  
	 ignore (printf "GT: findArray subgoal creation for %a\n" Satproc.dm msg)) ;
	for i = 1 to (Dag.dagSize () ) do 
	let node_i = Dag.findDag i in 
	if node_i.op = Dag.BFun("of") then (
		match (node_i.args) with
			(* strategy: consider every possible array we have seen so far *)
	  |	[a;b] when (b.op = Dag.Fun("array")) -> 
				(* is it contained inside our "ADDR" expression? *)
				let s_ADDR = List.assoc "ADDR" s in 
				let array_name = (match a.op with
					Dag.Var(v) -> v
				| Dag.Fun(v) -> v
				| _ -> E.s (E.bug "GT: findArray cannot handle %a" Dag.dn a)) in
				if (addrContainsArray s_ADDR array_name) then (
					(if (!debug) then  
						ignore (printf "GT: findArray guessing %a\n" Dag.dn node_i)) ;
					let s_RO = Subgoal.inputExpOfDagNode (List.nth b.args 0) in
					let s_T = Subgoal.inputExpOfDagNode (List.nth b.args 1) in
					let s_L = Subgoal.inputExpOfDagNode (List.nth b.args 2) in 
					let s_SIZET = sizeOfType s_T in 
					let s_A = Subgoal.inputExpOfDagNode a in
					(* ADDR = A+I, so I=ADDR-A *)
					let s_I = removeArrayFromAddr s_ADDR array_name in 
					let rec try_all offset tau = 
						(match tau with
							I.Fun("tuple2",[a;b]) -> 
								try_all offset a ; try_all (offset + numSizeOfType a) b 
						| I.Fun("tuple3",[a;b;c]) -> 
								try_all offset a ; try_all (offset + numSizeOfType a) b ; 
								try_all (offset + (numSizeOfType a) + (numSizeOfType b)) c 
						| I.Fun("tuple4",[a;b;c;d]) ->
								try_all offset a ; try_all (offset + numSizeOfType a) b ; 
								try_all (offset + (numSizeOfType a) + (numSizeOfType b)) c ; 
								try_all (offset + (numSizeOfType a) + (numSizeOfType b) 
																+ (numSizeOfType c)) d 
						| _ -> 
							let s_OFF = I.Int(offset) in 
							let s_TPRIME = tau in 
							let new_s = ("A",s_A) :: ("ADDR",s_ADDR) :: ("RO",s_RO) :: 
								("T",s_T) :: ("L",s_L) :: ("SIZET",s_SIZET) :: ("OFF",s_OFF) :: 
								("T'",s_TPRIME) :: ("I",s_I) :: s in
							did_something := true ;
							if (modCheck s_SIZET s_I s_OFF) then 
								Subgoal.add (makeSubGoal r msg new_s 1)
							else
							  if (!debug) then (
								ignore (printf "GT: failed modCheck %a %a %a\n" I.de s_SIZET
													I.de s_I I.de s_OFF))
					) in try_all 0 s_T 
				) else (
	(* ignore (printf "find: R/no %s %d = %a\n" array_name i Dag.dn node_i) ; *)
					() 
				) 
		| _ -> 
	(* ignore (printf "find: R/notArray %d = %a\n" i Dag.dn node_i) ; *)
				() 
	) done ;
	if (not(!did_something)) then (
	(if (!debug) then ignore (printf "findArray: no luck, trying ptropt\n"));
	for i = 1 to (Dag.dagSize () ) do 
	let node_i = Dag.findDag i in 
	if node_i.op = Dag.BFun("of") then (
		match (node_i.args) with
			(* strategy: consider every possible array we have seen so far *)
	  |	[a;b] when (b.op = Dag.Fun("ptropt")) & 
			((Subgoal.inputExpOfDagNode a) = (List.assoc "ADDR" s)) ->
			(if (!debug) then ignore (printf "findArray: valid ptropt!\n"));
			let s_ADDR = List.assoc "ADDR" s in 
			let s_RO = _rw in 
			let s_T = Subgoal.inputExpOfDagNode (List.nth b.args 0) in
			let s_TPRIME = s_T in
			let s_L = Subgoal.inputExpOfDagNode (List.nth b.args 1) in 
			let s_SIZET = sizeOfType s_T in 
			let s_A = Subgoal.inputExpOfDagNode a in
			let s_I = _num 0 in 
			let s_OFF = _num 0 in
			let new_s = ("A",s_A) :: ("ADDR",s_ADDR) :: ("RO",s_RO) :: 
			("T",s_T) :: ("L",s_L) :: ("SIZET",s_SIZET) :: ("OFF",s_OFF) :: 
			("T'",s_TPRIME) :: ("I",s_I) :: s in
			Subgoal.add (makeSubGoal r msg new_s 1)
		| _ -> () 
	)
	done
	) 
;;

(* 
 * Used to submit subgoals for expressions involving (sel4 M ADDR). 
 * Given that (= M (upd4 M ADDR' EXP)), we know that (= OTHER (sel4 M ADDR)
 * for some OTHER: only 2 choices from McCarthy's axioms. 
 *)
let findMemory r msg s =
	(if (!debug) then  
	 ignore (printf "GT: findMemory subgoal creation for %a\n" Satproc.dm msg)) ;
	let node_M = Dag.internExp (List.assoc "M" s) in
	let rec tryEq ml = 
	 (match ml with
		m::rest -> 
			let node_m = Dag.findDag m in (* node_i == M in EDAG *)
			if (node_m.op = Dag.Fun("upd4")) then (
				(if (!debug) then
					ignore (printf "GT: findMemory guessing %a\n" Dag.dn node_m)) ;
				let s_M2 = Subgoal.inputExpOfDagNode node_m in		
				let s_OTHER = Subgoal.inputExpOfDagNode (List.nth node_m.args 2) in
				(* only works if this upd4 addr corresponds to our sel4 addr *)
				let new_s = ("M2",s_M2) :: ("OTHER",s_OTHER) :: s in 
				Subgoal.add (makeSubGoal r msg new_s 1) ;
				(* now the case where the addrs are not equal *)
				let s_middle_M = Subgoal.inputExpOfDagNode (List.nth node_m.args 0) in
				(* bypass this memory update *)
				let s_OTHER = (I.Fun("sel4",[s_middle_M;(__"ADDR")])) in
				let new_s = ("M2",s_M2) :: ("OTHER",s_OTHER) :: s in 
				Subgoal.add (makeSubGoal r msg new_s 1)
			) else (
				tryEq rest (* this is not an upd4 expression, try again *)
			)
		| [] -> 
			(if (!debug) then
				ignore (printf "GT: findMemory not found: %a\n" I.de 
					(List.assoc "M" s))) ;
			()) in
	tryEq node_M.eqClass 
;; 

(*
 * specialSubgoalList is a list of rules that have special subgoal creation
 * functions. 
 *)
let specialSubgoalList = [
  { specialId = "bitOrAnd" ; makeSubgoals = findBitOr ; } ;
  { specialId = "safeRd" ;  makeSubgoals = findArray ; } ; 
  { specialId = "safeWr" ;  makeSubgoals = findArray ; } ; 
  { specialId = "ofSel"  ;  makeSubgoals = findArray ; } ; 
  { specialId = "geqSel" ;  makeSubgoals = findMemory ; } ; 
] ; (* end of : specialSubgoalList *)

	
(* GT maintains an (inefficient, useless) undoStack, ala engine.ml *)
type undoElem = 
	  Snapshot (* marks a snapshot *)

let undoStack : undoElem list ref = ref [] 

(* counter for making unique message ids *)
let counter = ref 0
let prfId () = 
	counter := !counter + 1 ;
	!counter

(* 
 * Record a snapshot 
 *)
let snapshot () = 
	undoStack := Snapshot :: !undoStack

(* 
 * Discard all assertions since last snapshot 
 *)
let undo () = 
	let rec loop = function
			Snapshot :: rest -> undoStack := rest
(* 		| _ :: rest -> loop rest *)
		| [] -> E.s (E.bug "GT tried to pop an empty stack (too many undos?)")
	in 
		loop !undoStack

(* 
 * Asserts a literal (the second argument). Accumulate the list of
 * equalities to the first argument  
 *
 * We filter out non-goals. If it is a goal, we check all of our rules
 * against it (unification). For each once that matches: (1) if it has
 * no hypotheses, we prove it now or (2) if it has hypotheses, we determine
 * the appropriate subgoal (via substitution) and then add it to the
 * subgoal manager. 
 *
 * Input Params: 
 * 	ml : message list			  Accumulated equalities
 * 	msg : message						Asserted literal
 * 	retval : message list   New accumulated equalities
 *
 *  message = { body:msgbody , owner:ident , id:int, goal:bool } 
 *)
let assume ml msg = 
   if (!debug) then ignore 
     (printf "GT: assume %a\n" S.dm msg) ;
   if (not msg.S.goal) then (
		(* we only care about goals! *)
     ml
   ) else ( (* this is a goal *)
	  (* check all of our general tacticals to see if they are applicable *)
     let msgDagNode = 
			match msg.S.body with
				Satproc.Literal(dn) -> dn
			| Satproc.Equality(d1,d2) -> (Dag.newNode Dag.Eq (d1::[d2]))
		in 
		let rec checkAll tlist =
			match tlist with 
			  [] -> ml
			| tactical :: rest -> 
				(try (
					if (!debug) then ignore 
						(printf "GT: trying to match rule %s\n" tactical.ident) ; 
				  let s = matches tactical.conc msgDagNode in
					(* if we didn't throw an exception by now, it matches! *)
					if (!debug) then ignore 
						(printf "GT: matches rule %s\n" tactical.ident) ;
					match tactical.hyp with
					  [] -> (* we can just prove it! *)
						let my_sub_exp_fun a = (Proof.E((subGoalOfExp a s))) in 
						if (!debug) then ignore 
							(printf "GT: proving Goal NOW via %s\n" tactical.ident) ;
						raise (S.Contradiction(P.G("not_FI",(P.P(P.G(tactical.ident,
							(List.map my_sub_exp_fun tactical.exp))))::
										[ (P.P(!Satproc.proveIt msg))] )))
                                        | _ -> 
						let rec try_special l = 
							match l with
								[] -> Subgoal.add (makeSubGoal tactical msg s 0)
							| spec::b -> if (spec.specialId = tactical.ident) then
									spec.makeSubgoals tactical msg s
									else try_special b
						in try_special specialSubgoalList
				) with No_Match -> 
					if (!debug && !rule_debug) then ignore 
						(printf "GT: failed to match rule %s\n" tactical.ident) ;
				  ())
				; 
				checkAll rest
		in
			checkAll ruleList ; 
			ml 
	)
;;

(*
 * Prove a message generated earlier. 
 * 
 *	msg : message
 *	retval : P.proof
 *)
let proveIt msg = E.s (E.unimp "GT.proveIt")
    
(* 
 * Called to initialize the decision procedure and to create a table of 
 * the entry points. 
 *
 * 	id : ident
 * 	retval : entries
 *)
let initialize id = 
  gtid := id;
  { S.snapshot = snapshot;
    S.undo = undo;
    S.assume = assume;
    S.proveIt = proveIt;
    S.debug = debug;
    S.ident  = id;
    S.reinitialize = function () -> ();
   } 
(* eof: tactical.ml *)
